#include <stdio.h>

#include <velib/velib_inc_mk2.h>

#include <velib/base/base.h>
#include <velib/mk2/hal.h>
#include <velib/mk2/mk2_update.h>
#include <velib/mk2/vebus_device.h>
#include <velib/platform/console.h>
#include <velib/utils/ve_assert.h>
#include <velib/utils/ve_logger.h>
#include <velib/utils/ve_timer.h>

static un8 startupTimer;
struct Mk2S mk2;

extern int forceUpdate;
extern int noTests;

typedef enum {
	ST_RESULT_FAIL_BUS_FREE_LINE,
	ST_RESULT_FAIL_VEBUS_DRIVER_ENABLE,
	ST_RESULT_FAIL_VEBUS_TX,
	ST_RESULT_FAIL_STANDBY_LINE,
	ST_RESULT_FAIL_UNKNOWN_CAUSE
} SelfTestResult;

void mk2FhRawFrameReceivedEvent(Mk2MsgRx *mk2msg)
{
	mk2Dump("<", mk2msg);
}

void mk2FhFrameSendEvent(Mk2MsgTx *mk2msg)
{
	//mk2Dump(">", (Mk2MsgRx *) mk2msg);
}

void mk2FhFrameReceivedEvent(Mk2MsgRx *mk2msg)
{
	VE_UNUSED(mk2msg);
}

static void doExit(int ret)
{
	mk2HoldInReset(veTrue);
	mk2EnsureVebusPower(veFalse);
	pltExit(ret);
}

static void done(void)
{
	printf("OK\r\n");
	doExit(0);
}

static void error(SelfTestResult result, un32 additionalInfo)
{
	switch (result)
	{
	case ST_RESULT_FAIL_BUS_FREE_LINE:
		fprintf(stderr, "ERROR: vebus line free %u\n", additionalInfo);
		break;
	case ST_RESULT_FAIL_VEBUS_DRIVER_ENABLE:
		fprintf(stderr, "ERROR: driver enable %u\n", additionalInfo);
		break;
	case ST_RESULT_FAIL_VEBUS_TX:
		fprintf(stderr, "ERROR: vebus tx %u\n", additionalInfo);
		if (additionalInfo & 4)
			fprintf(stderr, "VVC tester not connected?\n");
		break;
	case ST_RESULT_FAIL_STANDBY_LINE:
		fprintf(stderr, "ERROR: standby line %u\n", additionalInfo);
		break;
	case ST_RESULT_FAIL_UNKNOWN_CAUSE:
		fprintf(stderr, "ERROR: unknown %u\n", additionalInfo);
		break;
	}

	doExit(1);
}

static void mk2UpdateDone(Mk2Result result)
{
	if (result != MK2_OK) {
		fprintf(stderr, "ERROR: mk2 update failed\n");
		doExit(1);
	}

	startupTimer = 20;
	mk2HoldInReset(veTrue);
	mk2EnsureVebusPower(veTrue);
}

static Mk2EventReply testStandbyLineMessageHandler(Mk2Event mk2event, Mk2MsgRx *msg)
{
	if (mk2event == MK2_FRAME && msg->buf[0] == 0xFF &&
			msg->buf[1] == 'W' && msg->length >= 5)
	{
		if ((msg->buf[4] & 0x80))
			error(ST_RESULT_FAIL_STANDBY_LINE, 1);

		return MK2_DONE;
	}

	return MK2_CONTINUE;
}

static veBool startStandbyLineTest(Mk2Callback* callback)
{
	Mk2MsgTx* msg = mk2Msg('W');
	if (!msg) {
		veAssert(veFalse);
		return veFalse;
	}

	mk2EnsureVebusPower(veFalse);

	msg->length = 5;
	msg->buf[2] = WM_CMD_SEND_SOFTWARE_VERSION_PART1;
	msg->buf[3] = 0x34;
	msg->buf[4] = 0x12;
	msg->handler = testStandbyLineMessageHandler;
	msg->callback = callback;

	logI("selftest", "checking standby (released)");
	mk2FhQueueFrame(msg);

	return veTrue;
}

static void mk2SelfTestDone(void)
{
	startStandbyLineTest(done);
}

/*
 * The vcc tester sents the W message back for the selftest,
 * but the mk2 will pass it on. Parse it and check the standby line.
 */
static Mk2EventReply testBounceMessageHandler(Mk2Event mk2event, Mk2MsgRx *msg)
{
	if (mk2event == MK2_FRAME && msg->buf[0] == 0xFF &&
			msg->buf[1] == 'W' && msg->length >= 5)
	{
		if ((msg->buf[4] & 0x80) == 0)
			error(ST_RESULT_FAIL_STANDBY_LINE, 0);

		return MK2_DONE;
	}

	return MK2_CONTINUE;
}

/* The expected values to be returned by the MK2 after a self-test */
#define MK2_TEST_RESULT0_MIN					80
#define MK2_TEST_RESULT0_MAX					119

#define MK2_TEST_RESULT1_MASK_RS485_DISABLE		0x01
#define MK2_TEST_RESULT1_MASK_TX_ERROR			0x06

/* parse the selftest response */
static Mk2EventReply selftestHandler(Mk2Event mk2event, Mk2MsgRx *msg)
{
	if (mk2event == MK2_FRAME && msg->buf[0] == 0xFF &&
			msg->buf[1] == 'T' && msg->length >= 4)
	{
		un8 result0 = msg->buf[2];
		un8 result1 = msg->buf[3];

		/* The rise time of the bus free line is outside the permitted range */
		if (result0 < MK2_TEST_RESULT0_MIN || result0 > MK2_TEST_RESULT0_MAX)
			error(ST_RESULT_FAIL_BUS_FREE_LINE, result0);

		/* The RS485 driver could not be disabled */
		if (result1 & MK2_TEST_RESULT1_MASK_RS485_DISABLE)
			error(ST_RESULT_FAIL_VEBUS_DRIVER_ENABLE, result1);

		/* There was a problem driving either a hard or a soft 1 */
		if (result1 & MK2_TEST_RESULT1_MASK_TX_ERROR)
			error(ST_RESULT_FAIL_VEBUS_TX, result1);

		/*
		 * The MK2 returned a non-zero result1 that did not correspond to any
		 * of the known error codes.
		 */
		if (result1)
			error(ST_RESULT_FAIL_UNKNOWN_CAUSE, result1);

		/* not done yet, the W resonse from the vvc tester should follow */
		mk2.fh.msgOut->handler = testBounceMessageHandler;
	}

	if (mk2event == MK2_TIMEOUT) {
		printf("ERROR: no replty to selftest command\n");
		doExit(1);
	}

	return MK2_CONTINUE;
}

/*
 * The W frame needs time to get forwarded to the vvc tester, since
 * the test command will stop the tranmission. So delay after the W
 * message is sent.
 */
static Mk2EventReply delayHandler(Mk2Event mk2event, Mk2MsgRx *msg)
{
	if (mk2event == MK2_TX_SENT) {
		veWait(100000);
		return MK2_DONE;
	}

	return MK2_CONTINUE;
}

/*
 * The mk2 selftest expects a byte to be received from the vebus side
 * at 1200 baud. This is done by sending a msg to the tester, which
 * will be bounced after 200ms. So a msg is send, and after 100ms the
 * selftest is initiated, which should receive the msg after 100ms or so.
 */
veBool startMk2SelfTest(Mk2Callback* callback)
{
	Mk2MsgTx* msg = mk2Msg('W');
	if (!msg) {
		veAssert(veFalse);
		return veFalse;
	}

	msg->length = 5;
	msg->buf[2] = WM_CMD_SEND_SOFTWARE_VERSION_PART1;
	msg->buf[3] = 0x34;
	msg->buf[4] = 0x12;
	msg->handler = delayHandler;

	logI("selftest", "sending dummy frame to vvc tester");
	mk2FhQueueFrame(msg);

	/* queue the selftest request for the mk2 */
	msg = mk2Msg('T');
	if (!msg) {
		veAssert(veFalse);
		return veFalse;
	}

	msg->handler = selftestHandler;
	msg->callback = callback;
	mk2FhQueueFrame(msg);

	return veTrue;
}

static void setHardwareIdDone(void)
{
	startMk2SelfTest(mk2SelfTestDone);
}

static Mk2EventReply setHardwareIdHandler(Mk2Event mk2event, Mk2MsgRx *msg)
{
	if (mk2event == MK2_FRAME && msg->buf[0] == 0xFF &&
			msg->buf[1] == 'I' && msg->length > 4)
	{
		return MK2_DONE;
	}

	return MK2_CONTINUE;
}

veBool setHardwareId(Mk2Callback* callback)
{
	Mk2MsgTx* msg = mk2Msg('I');

	msg->length = 4;
	msg->buf[2] = 0x01; /* first 0x01 indicates to only write hardware id, the value follows */
	msg->buf[3] = 0x01;
	msg->handler = setHardwareIdHandler;
	msg->callback = callback;
	mk2FhQueueFrame(msg);

	return veTrue;
}


/* 1< mk2 is there, set its state or update when outdated */
static void mk2GetVersionDone(void)
{
	if ((mk2.version % 1000) != mk2UpdateVersion || forceUpdate) {
		if (forceUpdate) {
			forceUpdate = veFalse;
			logI("mk2", "forcing update from version %d to %d", mk2.version % 1000, mk2UpdateVersion);
		} else {
			logI("mk2", "wrong version %d, want %d, starting update", mk2.version % 1000, mk2UpdateVersion);
		}
		mk2FirmwareUpgrade(mk2UpdateDone);
		return;
	}

	if (noTests) {
		logI("mk2", "not running test on request!");
		doExit(0);
	}

	setHardwareId(setHardwareIdDone);
}


/*
 * This is called early, there is nothing know yet about dev.
 * (besides its short address)
 */
void taskInit(void)
{
	mk2FhInit();
	mk2Init();

	/*
	 * Setup the hardware, these are two distinct calls since the bpp3 does
	 * use the regular pc serial interface, but has hardware to control some
	 * mk2 io pins.
	 */
	mk2HalInit();
	mk2IoInit();

	mk2BreakClear();
	startupTimer = 10;
	mk2HoldInReset(veTrue);
	mk2EnsureVebusPower(veTrue); /* power vebus */

	pltInterruptEnable();
}

void taskUpdate(void)
{
	mk2FhUpdate();
}

/* 1> first get the version of the mk2 */
static void mk2Begin(void)
{
	mk2.version = 0;
	logI("mk2", "getversion");
	mk2GetVersion(&mk2.version, mk2GetVersionDone);
}

void taskTick(void)
{
	if (mk2UpdateBusy()) {
		mk2FirmwareTick();
		return;
	}

	mk2FhTick();

	/*
	 * Get the mk2 out of reset before the startupTimer expires, so it has
	 * some time to startup.
	 */
	if (startupTimer == 5)
		mk2HoldInReset(veFalse);

	/* begin the startup procedure */
	if (startupTimer != 0 && --startupTimer==0)
		mk2Begin();

}

/** No response is received after several tries. */
void mk2NoResponseCallback(void)
{
	logI("mk2", "timeout");

	/* if the mk2 didn't respond, try to update it.. */
	if (!mk2.version) {
		mk2FirmwareUpgrade(mk2UpdateDone);
		return;
	}

	fprintf(stderr, "ERROR: mk2 did not respond\n");
	pltExit(1);
}

veBool vebusCanBeControlled(void)
{
	return veTrue;
}
